import unittest
import pkg_resources

from . import BaseTestCase
from ..client import Client
from ..orm.object import Object
from ..orm.record import Record
from ..service.service import ServiceManager
from ..plugin import Plugin

VERSION_CLASSES = (pkg_resources.SetuptoolsLegacyVersion,
                   pkg_resources.SetuptoolsVersion)


class Test_10_Client(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)

    def test_120_username(self):
        self.assertEqual(self.client.username, self.env.user)
        self.assertIsInstance(self.client.user, Record)
        self.assertEqual(self.client.user.login, self.env.user)

    def test_122_user_context(self):
        uctx = self.client.user_context
        self.assertDictEqual(
            uctx,
            self.client.get_obj('res.users').context_get())

    def test_125_server_version(self):
        # Check that server version is wrapped in parse_version. thus allows to
        # compare versions
        self.assertIsInstance(self.client.server_version, VERSION_CLASSES)

    def test_126_database_version_full(self):
        # Check that database full version is wrapped in parse_version.
        # thus allows to compare versions
        self.assertIsInstance(self.client.database_version_full,
                              VERSION_CLASSES)

    def test_127_database_version(self):
        # Check that database version is wrapped in parse_version.
        # thus allows to compare versions
        self.assertIsInstance(
            self.client.database_version, VERSION_CLASSES)

    def test_128_database_version_eq_server_version(self):
        self.assertEqual(self.client.server_version,
                         self.client.database_version)

    def test_130_get_obj(self):
        self.assertIn('res.partner', self.client.registered_objects)
        obj = self.client.get_obj('res.partner')
        self.assertIsInstance(obj, Object)

        # Check object access in dictionary style
        self.assertIs(obj, self.client['res.partner'])

    def test_142_get_obj_wrong(self):
        self.assertNotIn(
            'bad.object.name',
            self.client.registered_objects)
        with self.assertRaises(ValueError):
            self.client.get_obj('bad.object.name')

        with self.assertRaises(KeyError):
            self.client['bad.object.name']

    def test_150_to_url(self):
        url_tmpl = "%(protocol)s://%(user)s@%(host)s:%(port)s/%(dbname)s"
        cl_url = url_tmpl % self.env
        self.assertEqual(Client.to_url(self.client), cl_url)
        self.assertEqual(Client.to_url(self.env), cl_url)
        self.assertEqual(Client.to_url(None, **self.env), cl_url)
        self.assertEqual(self.client.get_url(), cl_url)

        with self.assertRaises(ValueError):
            Client.to_url('strange thing')

    def test_155_str(self):
        self.assertEqual(str(self.client), u"Client: %s" %
                         self.client.get_url())

    def test_155_repr(self):
        self.assertEqual(repr(self.client), str(self.client))

    def test_160_plugins(self):
        self.assertIn('Test', self.client.plugins.registered_plugins)
        self.assertIn('Test', self.client.plugins)
        self.assertIn(
            'Test', dir(
                self.client.plugins))  # for introspection
        # iteration over plugins names
        self.assertIn('Test', [p for p in self.client.plugins])
        self.assertIsInstance(self.client.plugins.Test, Plugin)
        self.assertIsInstance(self.client.plugins['Test'], Plugin)
        self.assertIs(
            self.client.plugins['Test'],
            self.client.plugins.Test)

        # check plugin's method result
        self.assertEqual(
            self.client.plugins.Test.test(),
            self.client.get_url())

        # Check plugin representation
        self.assertEqual(
            str(self.client.plugins),
            "odoo_rpc_client.plugin.PluginManager [%d]" % len(
                self.client.plugins))
        self.assertEqual(
            repr(self.client.plugins),
            "<odoo_rpc_client.plugin.PluginManager [%d]>" % len(
                self.client.plugins))
        self.assertEqual(
            str(self.client.plugins.Test),
            "odoo_rpc_client.plugin.Plugin:Test")
        self.assertEqual(
            repr(self.client.plugins.Test),
            "<odoo_rpc_client.plugin.Plugin:Test>")

    def test_162_plugins_wrong_name(self):
        self.assertNotIn(
            'Test_Bad',
            self.client.plugins.registered_plugins)
        self.assertNotIn('Test_Bad', self.client.plugins)
        self.assertNotIn(
            'Test_Bad', dir(
                self.client.plugins))  # for introspection

        with self.assertRaises(KeyError):
            self.client.plugins['Test_Bad']

        with self.assertRaises(AttributeError):
            self.client.plugins.Test_Bad

    def test_170_client_services(self):
        self.assertIsInstance(self.client.services, ServiceManager)
        self.assertIn('db', self.client.services)
        self.assertIn('object', self.client.services)
        self.assertIn('report', self.client.services)

        self.assertIn('db', self.client.services.service_list)
        self.assertIn('object', self.client.services.service_list)
        self.assertIn('report', self.client.services.service_list)

        self.assertIn('db', dir(self.client.services))
        self.assertIn('object', dir(self.client.services))
        self.assertIn('report', dir(self.client.services))

        # Test representations:
        self.assertEqual(
            str(self.client.services), u"ServiceManager for %s" %
            self.client)
        self.assertEqual(
            repr(self.client.services), u"<ServiceManager for %s>" %
            self.client)
        self.assertEqual(
            str(self.client.services.db), u"Service 'db' of %s" %
            self.client)
        self.assertEqual(
            repr(self.client.services.db), u"<Service 'db' of %s>" %
            self.client)
        self.assertEqual(
            str(self.client.services.report), u"Service 'report' of %s" %
            self.client)
        self.assertEqual(
            repr(self.client.services.report), u"<Service 'report' of %s>" %
            self.client)

        with self.assertRaises(AttributeError):
            self.client.services._private_service

    def test_180_execute_lt_v10(self):
        if self.client.server_version >= pkg_resources.parse_version('10.0'):
            raise unittest.SkipTest('Not applicable to Odoo 10.0')

        res = self.client.execute('res.partner', 'read', 1)
        self.assertIsInstance(res, dict)
        self.assertEqual(res['id'], 1)

        res = self.client.execute('res.partner', 'read', [1])
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0], dict)
        self.assertEqual(res[0]['id'], 1)

    def test_181_execute_gte_v10(self):
        if self.client.server_version < pkg_resources.parse_version('10.0'):
            raise unittest.SkipTest(
                'Not applicable to Odoo version less then 10.0')

        res = self.client.execute('res.partner', 'read', 1)
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0], dict)
        self.assertEqual(res[0]['id'], 1)

        res = self.client.execute('res.partner', 'read', [1])
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0], dict)
        self.assertEqual(res[0]['id'], 1)

    def test_190_not_equal(self):
        self.assertNotEqual(self.client, 42)

    def test_200_clean_caches(self):
        self.assertIsNotNone(self.client.user_context)
        self.assertIn('lang', self.client.user_context)

        self.client.clean_caches()

        self.assertIsNone(self.client._user)
        self.assertIsNotNone(self.client._username)
        self.assertIsNone(self.client._user_context)

    def test_210_from_url_host_port(self):
        cl = Client.from_url("%(host)s:%(port)s/" % self.env)
        self.assertIsInstance(cl, Client)
        self.assertEqual(cl.get_url(),
                         "xml-rpc://None@%(host)s:%(port)s/None" % self.env)

    def test_211_from_url_protocol_host_port_user_db(self):
        cl = Client.from_url(
            "%(protocol)s://%(user)s@%(host)s:%(port)s/%(dbname)s" % self.env)
        self.assertIsInstance(cl, Client)

    def test_212_from_url_protocol_host_port_user_db(self):
        cl = Client.from_url(
            "%(protocol)s://%(user)s:%(password)s@%(host)s:%(port)s/"
            "%(dbname)s" % self.env)
        self.assertIsInstance(cl, Client)
        self.assertEqual(cl.user.login, self.env['user'])

    def test_220_ref_existing(self):
        partner = self.client.ref('base.main_partner')
        self.assertIsInstance(partner, Record)
        self.assertEqual(partner._object.name, 'res.partner')

    def test_221_ref_un_existing(self):
        partner = self.client.ref('base.unexisting_partner_id')
        self.assertIsInstance(partner, bool)
        self.assertFalse(partner)

    def test_222_ref_no_module_spec(self):
        with self.assertRaises(ValueError):
            self.client.ref('main_partner')  # no module specified



from ..service.object import ObjectService
from .object import get_object

__all__ = ('Service',)


class Service(ObjectService):
    """ Service class to simplify interaction with 'object' service.
        Particulary, implements logic of choosing execute
        method ('execute' or 'execute_kw') to use.
        The last one cannot work with keyword arguments
    """

    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)
        self.__objects = {}   # cached objects

    def get_obj(self, object_name):
        """ Returns wraper around Odoo object 'object_name'
            which is instance of Object

            :param object_name: name of an object to get wraper for
            :type object_name: string
            :return: instance of Object which wraps choosen object
            :rtype: Object
        """
        if object_name in self.__objects:
            return self.__objects[object_name]

        if object_name not in self.get_registered_objects():
            raise ValueError("There is no object named '%s'" % object_name)

        obj = get_object(self, object_name)
        self.__objects[object_name] = obj
        return obj

    def clean_cache(self):
        """ Cleans caches, to fill them with fresh data
            on next call of related methods
        """
        super(Service, self).clean_cache()
        self.__objects = {}



from . import (object,  # noqa
               report,  # noqa
               db)      # noqa
from .service import (get_service_class,  # noqa
                      ServiceBase,        # noqa
                      ServiceManager)     # noqa

__all__ = (
    'get_service_class',
    'ServiceBase',
    'ServiceManager')



from ..plugin import Plugin
from ..orm.object import Object
from ..utils import (stdcall,
                     DirMixIn)


# Overriden to add shortcut for module update
class ModuleObject(Object):
    """ Add shortcut methods to 'ir.module.module' object / model
        to install or upgrade modules

        Also this methods will be available for Record instances too
    """

    # Apply this extension only to 'ir.module.module' Object / Model
    class Meta:
        name = 'ir.module.module'

    @stdcall
    def upgrade(self, ids, context=None):
        """ Immediatly upgrades module
        """
        kwargs = {} if context is None else {'context': context}
        res = self.button_immediate_upgrade(ids, **kwargs)
        # because new models may appear in DB,
        # so registered_objects shoud be refreshed
        self.client.clean_caches()
        return res

    @stdcall
    def install(self, ids, context=None):
        """ Immediatly install module
        """
        kwargs = {} if context is None else {'context': context}
        res = self.button_immediate_install(ids, **kwargs)
        # because new models may appear in DB,
        # so registered_objects shoud be refreshed
        self.client.clean_caches()
        return res


class ModuleUtils(Plugin, DirMixIn):
    """ Utility plugin to simplify module management

        Allows to access Odoo module objects as attributes of this plugin:

        .. code:: python

            # this method supports IPython autocomplete
            db.plugins.module_utils.m_stock

        or dictionary style access to modules:

        .. code:: python

            db.plugins.moduld_utils['stock']

        which is equivalent to

        .. code:: python

            db.get_obj('ir.module.module').search_records(
                [('name','=','stock')])[0]

        Also autocomplete in IPython supported for this syntax
    """

    class Meta:
        name = "module_utils"

    def __init__(self, *args, **kwargs):
        super(ModuleUtils, self).__init__(*args, **kwargs)
        self._modules = None

    @property
    def modules(self):
        """ Returns dictionary of modules registered in system.

            Result dict is like: ``{'module_name': module_inst}``

            where *module_inst* is *Record* instance for this module
        """
        if self._modules is None:
            modules = self.client['ir.module.module'].search_records([])
            self._modules = {m.name: m
                             for m in modules}
        return self._modules

    @property
    def installed_modules(self):
        """ RecordList with list of modules installed in currenct database

            :rtype: RecordList
        """
        return self.client['ir.module.module'].search_records(
            [('state', '=', 'installed')])

    def update_module_list(self):
        """ Update module list

            If there are some modules added to server,
            update list, to be able to installe them.
        """
        self._modules = None
        return self.client['ir.module.module'].update_list()

    def __dir__(self):
        res = super(ModuleUtils, self).__dir__()
        res.extend(['m_' + i for i in self.modules.keys()])
        return res

    def __getitem__(self, name):
        return self.modules[name]

    def __getattr__(self, name):
        if name.startswith('m_') and name[2:] in self.modules:
            return self.modules[name[2:]]
        raise AttributeError("No attribute %s in object %s" % (name, self))

    def __contains__(self, name):
        return name in self.modules



import six
import unittest
from ..utils import AttrDict
from ..connection.local import ConnectorLocal

import logging
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)


# Python 2/3 compatability
try:
    import unittest.mock as mock
except ImportError:
    import mock

__all__ = ('BaseTestCase', 'mock')


class ConnectorLocalTestExt(ConnectorLocal):
    """ Extend local connector to get odoo options from environment
    """

    def _start_odoo_services(self):
        if not self.odoo_args:
            import os
            self.odoo_args = eval(
                os.environ.get('ODOO_TEST_LOCAL_ARGS', '[]'))
        return super(ConnectorLocalTestExt, self)._start_odoo_services()


class BaseTestCase(unittest.TestCase):
    """Instanciates an ``odoorpc.ODOO`` object, nothing more."""
    def setUp(self):
        import os
        try:
            port = int(os.environ.get('ODOO_TEST_PORT', 8069))
        except ValueError:
            raise ValueError("The port must be an integer")
        self.env = AttrDict({
            'protocol': os.environ.get('ODOO_TEST_PROTOCOL', 'xml-rpc'),
            'host': os.environ.get('ODOO_TEST_HOST', 'localhost'),
            'port': port,
            'dbname': os.environ.get('ODOO_TEST_DB',
                                     'odoo_rpc_client_test_db'),
            'user': os.environ.get('ODOO_TEST_USER', 'admin'),
            'password': os.environ.get('ODOO_TEST_PASSWORD', 'admin'),
            'super_password': os.environ.get('ODOO_TEST_SUPER_PASSWORD',
                                             'admin'),
            'local_args': eval(os.environ.get('ODOO_TEST_LOCAL_ARGS', 'None'))
        })

        self.test_db_service = os.environ.get('TEST_DB_SERVICE', False)
        self.recreate_db = os.environ.get('RECREATE_DB', False)

    if six.PY3:
        def assertItemsEqual(self, *args, **kwargs):
            return self.assertCountEqual(*args, **kwargs)



#!/usr/bin/env python

import os.path
import sys
from setuptools import setup

# workaround for: https://github.com/travis-ci/travis-ci/issues/1778
import multiprocessing  # noqa

# load version info
version_file = os.path.join(os.path.dirname(__file__),
                            'odoo_rpc_client',
                            'version.py')
readme_file = os.path.join(os.path.dirname(__file__), 'README.rst')

if sys.version_info < (3,):
    execfile(version_file)
else:
    with open(version_file, 'rb') as f:
        exec(compile(f.read(), version_file, 'exec'))


setup(name='odoo_rpc_client',
      version=version,
      description='Odoo/OpenERP library for RPC',
      author='Dmytro Katyukha',
      author_email='firemage.dima@gmail.com',
      url='https://github.com/katyukha/odoo-rpc-client',
      long_description=open(readme_file).read(),
      packages=['odoo_rpc_client',
                'odoo_rpc_client.connection',
                'odoo_rpc_client.service',
                'odoo_rpc_client.orm',
                'odoo_rpc_client.tests',
                'odoo_rpc_client.plugins',
                ],
      license="GPL",
      classifiers=[
          'Development Status :: 3 - Alpha',
          # 'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Utilities',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords=['odoo', 'odoo-rpc', 'rpc', 'xmlrpc',
                'xml-rpc', 'json-rpc', 'jsonrpc', 'odoo-client', 'openerp'],
      extras_require={
          'all': ['anyfield'],
      },
      install_requires=[
          'six>=1.10',
          'extend_me>=1.1.3',
          'setuptools>=18',
          'requests>=2.7',
          'simplejson',
      ],
      tests_require=[
          'mock',
          'coverage',
          'anyfield',
      ],
      test_suite='odoo_rpc_client.tests.all',
)



from pkg_resources import parse_version

from ..service.service import ServiceBase


class ObjectService(ServiceBase):
    """ Service class to simplify interaction with 'object' service
        Particulary, implements logic of choosing execute method
        ('execute' or 'execute_kw') The last one cannot work with
        keyword arguments(
    """
    class Meta:
        name = 'object'

    def __init__(self, *args, **kwargs):
        super(ObjectService, self).__init__(*args, **kwargs)
        self._registered_objects = None

    def execute(self, obj, method, *args, **kwargs):
        """First arguments should be 'object' and 'method' and next
           will be passed to method of given object
        """
        # avoid sending context when it is set to None
        # because of it is problem of xmlrpc
        if 'context' in kwargs and kwargs['context'] is None:
            kwargs = kwargs.copy()
            del kwargs['context']

        result = self._service.execute_kw(self.client.dbname,
                                          self.client.uid,
                                          self.client._pwd,
                                          obj,
                                          method,
                                          args,
                                          kwargs)
        return result

    def execute_wkf(self, object_name, signal, object_id):
        """ Triggers workflow event on specified object

            :param str object_name: name of object/model to trigger workflow on
            :param str signal: name of signal to send to workflow
            :param int object_id: ID of document (record) to send signal to
        """
        result_wkf = self._service.exec_workflow(self.client.dbname,
                                                 self.client.uid,
                                                 self.client._pwd,
                                                 object_name,
                                                 signal,
                                                 object_id)
        return result_wkf

    def _get_registered_objects(self):
        """ Implementation of get registered models (objects)
            Could be overridden by extensions
        """
        if self.client.server_version > parse_version('8.0'):
            read = self.execute('ir.model', 'search_read',
                                domain=[], fields=['model'])
        else:
            ids = self.execute('ir.model', 'search', [])
            read = self.execute('ir.model', 'read', ids, ['model'])
        return [x['model'] for x in read]

    def get_registered_objects(self):
        """ Returns list of registered objects in database
        """
        if self._registered_objects is not None:
            return self._registered_objects
        self._registered_objects = self._get_registered_objects()
        return self._registered_objects

    def clean_cache(self):
        """ Cleans service cache, to fill them with fresh data
            on next call of related methods
        """
        self._registered_objects = None



import six
from extend_me import (ExtensibleByHashType,
                       Extensible)

from ..utils import DirMixIn


__all__ = ('get_service_class', 'ServiceBase', 'ServiceManager')


@six.python_2_unicode_compatible
class ServiceManager(Extensible, DirMixIn):
    """ Class to hold services related to specific client and to
        automaticaly clean service cached on update of service classes

        Usage::

            services = ServiceManager(client)
            services.service_list          # get list of registered services
            services.object                # returns service with name 'object'
            services['common']             # returns service with name 'common'
            services.get_service('report') # returns service named 'report'
    """

    __managers = []

    def __new__(cls, *args, **kwargs):
        inst = super(ServiceManager, cls).__new__(cls, *args, **kwargs)
        cls.__managers.append(inst)
        return inst

    def __init__(self, client):
        self._client = client
        self.__services = {}

    def __dir__(self):
        res = set(super(ServiceManager, self).__dir__())
        res.update(self.service_list)
        return list(res)

    @property
    def client(self):
        """ Client instance this ServiceManager is bounded to
        """
        return self._client

    @property
    def service_list(self):
        """ Returns list of all registered services
        """
        service_names = set()
        service_names.update(list(self.__services))
        service_names.update(ServiceType.get_registered_names())
        return list(service_names)

    def get_service(self, name):
        """ Returns instance of service with specified name

            :param name: name of service
            :return: specified service instance
        """
        service = self.__services.get(name, None)
        if service is None:
            cls = get_service_class(name)
            srv = self._client.connection.get_service(name)
            service = cls(srv, self.client, name)
            self.__services[name] = service
        return service

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError("Service '%s' not found."
                                 "Service names could not be "
                                 "started with '_'.")
        return self.get_service(name)

    def __getitem__(self, name):
        return self.get_service(name)

    def __contains__(self, name):
        return name in self.service_list

    @classmethod
    def clean_caches(cls):
        """ Cleans saved service instances, so on next access
            new service instances will be generated.
            This usualy happens when new service extension enabled
            (new class inherited from ServiceBase created)
        """
        for manager in cls.__managers:
            manager.clean_cache()

    def clean_cache(self):
        """ Cleans manager's service cache.
        """
        self.__services = {}

    def clean_service_caches(self):
        """ Clean caches of all services handled by this mananger
            usualy this should be called on module update,
            when list of available objects or reports changed
        """
        for service in self.__services.values():
            service.clean_cache()

    def __str__(self):
        return u"ServiceManager for %s" % self.client

    def __repr__(self):
        return u"<%s>" % str(self)


class ServiceMetaMixIn(type):
    """ Simple meta mixin class, that cleans all service caches,
        when new service class created / imported
    """
    def __new__(mcs, *args, **kwargs):
        cls = super(ServiceMetaMixIn, mcs).__new__(mcs, *args, **kwargs)

        # Clean caches only if new user extension class defined
        # This condition is required, because extension mechanism creates new
        # class, that is subclass of all classes defined by user, and it marked
        # by attribute '_generated', so we do not need to clean caches when
        # this class is created. Such classes will be created each time,
        # service instance is accessed first time. For example, when connecting
        # we use 'common' service to login to database, and when we access this
        # service first time, new *_generated* service class created, when next
        # we access 'object' service, then again new *_generated* class is
        # created for this 'object' service. But no new user defined classes
        # created, thus it is possibly leads to dubling of some rpc requests,
        # such as 'registered_objects'.
        if not getattr(cls, '_generated', False):
            ServiceManager.clean_caches()

        return cls


ServiceType = ExtensibleByHashType._('Service',
                                     hashattr='name',
                                     with_meta=ServiceMetaMixIn)


def get_service_class(name):
    """ Return service class specified by it's name
    """
    return ServiceType.get_class(name, default=True)


@six.python_2_unicode_compatible
class ServiceBase(six.with_metaclass(ServiceType, object)):
    """ Base class for all Services

        :param service: instance of original service class.
                        must support folowing syntax
                        ``service.service_method(args)``
                        to call remote methods
        :param client: instance of Client, this service is binded to
    """

    def __init__(self, service, client, name):
        self._client = client
        self._service = service
        self._name = name

    @property
    def client(self):
        """ Related Client instance
        """
        return self._client

    @property
    def name(self):
        """ Service name
        """
        return self._name

    def __getattr__(self, name):
        return getattr(self._service, name)

    def clean_cache(self):
        """ To be implemented by subclasses, if needed
        """
        pass

    def __str__(self):
        return "Service '%s' of %s" % (self.name, self.client)

    def __repr__(self):
        return u"<%s>" % str(self)



# python imports
from six.moves import xmlrpc_client as xmlrpclib

# project imports
from .connection import ConnectorBase
from ..utils import ustr
from .. import exceptions as exceptions


class XMLRPCError(exceptions.ConnectorError):
    """ Exception raised on XMLRpc errors

        :param xmlrpclib.Fault fault_instance: exception raised by XMLRPC lib
    """
    def __init__(self, fault_instance):
        self._fault = fault_instance
        msg = (u"A fault occured\n"
               u"Fault code: %s\n"
               u"Fault string: %s\n"
               u"" % (ustr(fault_instance.faultCode),
                      ustr(fault_instance.faultString)))
        msg = msg.encode('utf-8')
        super(XMLRPCError, self).__init__(msg)

    @property
    def fault(self):
        """ Return xmlrpclib.Fault instance related to this error
        """
        return self._fault


class XMLRPCMethod(object):
    """ Class wrapper around XML-RPC method to wrap xmlrpclib.Fault
        into XMLRPCProxy
    """

    def __init__(self, method):
        self.__method = method

    def __getattr__(self, name):  # pragma: no cover
        return XMLRPCMethod(getattr(self.__method, name))

    def __call__(self, *args):
        try:
            res = self.__method(*args)
        except xmlrpclib.Fault as fault:
            raise XMLRPCError(fault)
        return res


class XMLRPCProxy(xmlrpclib.ServerProxy):
    """ Wrapper class around XML-RPC's ServerProxy to wrap method's errors
        into XMLRPCError class
    """
    def __getattr__(self, name):
        res = xmlrpclib.ServerProxy.__getattr__(self, name)
        if isinstance(res, xmlrpclib._Method):
            res = XMLRPCMethod(res)
        return res


class ConnectorXMLRPC(ConnectorBase):
    """ XML-RPC connector

        Note: extra_arguments may be same as parametrs of xmlrpclib.ServerProxy
    """
    class Meta:
        name = 'xml-rpc'
        ssl = False

    def get_service_url(self, service_name):
        addr = self.host
        if self.port not in (None, 80):
            addr += ':%s' % self.port
        proto = 'https' if self.Meta.ssl else 'http'
        return '%s://%s/xmlrpc/%s' % (proto, addr, service_name)

    def _get_service(self, name):
        return XMLRPCProxy(self.get_service_url(name), **self.extra_args)


class ConnectorXMLRPCS(ConnectorXMLRPC):
    """ XML-RPCS Connector

        Note: extra_arguments may be same as parametrs of xmlrpclib.ServerProxy
    """
    class Meta:
        name = 'xml-rpcs'
        ssl = True



# Python imports
import six
import extend_me

from .utils import DirMixIn

PluginMeta = extend_me.ExtensibleByHashType._('Plugin', hashattr='name')


@six.python_2_unicode_compatible
class Plugin(six.with_metaclass(PluginMeta)):
    """ Base class for all plugins, extensible by name

        (uses metaclass extend_me.ExtensibleByHashType)

        :param client: instance of Client to bind plugins to
        :type client: odoo_rpc_client.client.Client instance

        Example of simple plugin::

            from odoo_rpc_client.plugin import Plugin

            class AttandanceUtils(Plugin):

                # This is required to register Your plugin
                # *name* - is for db.plugins.<name>
                class Meta:
                    name = "attendance"

                def get_sign_state(self):
                    # Note: folowing code works on version 6 of Openerp/Odoo
                    emp_obj = self.client['hr.employee']
                    emp_id = emp_obj.search(
                        [('user_id', '=', self.client.uid)])
                    emp = emp_obj.read(emp_id, ['state'])
                    return emp[0]['state']

        This plugin will automaticaly register itself in system,
        when module which contains it will be imported.
    """

    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        """ Related Client instance
        """
        return self._client

    def __str__(self):
        return u"odoo_rpc_client.plugin.Plugin:%s" % self.Meta.name

    def __repr__(self):
        return u"<%s>" % (str(self))


class TestPlugin(Plugin):
    """ Jusn an example plugin to test if plugin logic works
    """

    class Meta:
        name = 'Test'

    def test(self):
        return self.client.get_url()


@six.python_2_unicode_compatible
class PluginManager(extend_me.Extensible, DirMixIn):
    """ Class that holds information about all plugins

        :param client: instance of Client to bind plugins to
        :type client: odoo_rpc_client.client.Client instance

        Plugiins will be accessible via index or attribute syntax::

            plugins = PluginManager(client)
            plugins.Test   # acceps plugin 'Test' as attribute
            plugins['Test']  # access plugin 'Test' via indexing
    """
    def __init__(self, client):
        """
        """
        self.__client = client
        self.__plugins = {}

    def __getitem__(self, name):
        plugin = self.__plugins.get(name, False)
        if plugin is False:
            try:
                pluginCls = type(Plugin).get_class(name)
            except ValueError as e:
                raise KeyError(str(e))

            plugin = pluginCls(self.__client)
            self.__plugins[name] = plugin
        return plugin

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __contains__(self, name):
        return name in self.registered_plugins

    def __iter__(self):
        return iter(self.registered_plugins)

    def __len__(self):
        return len(self.registered_plugins)

    def __dir__(self):
        res = super(PluginManager, self).__dir__()
        res.extend(self.registered_plugins)
        return res

    @property
    def registered_plugins(self):
        """ List of names of registered plugins
        """
        return type(Plugin).get_registered_names()

    def refresh(self):
        """ Clean-up plugin cache
            This will force to reinitialize each plugin when asked
        """
        self.__plugins = {}
        return self

    def __str__(self):
        return u"odoo_rpc_client.plugin.PluginManager [%d]" % len(self)

    def __repr__(self):
        return u"<%s>" % str(self)
